@@ -1,5 +1,5 @@
 /*
- * The ARP Scanner (arp-scan) is Copyright (C) 2005-2007 Roy Hills,
+ * The ARP Scanner (arp-scan) is Copyright (C) 2005-2008 Roy Hills,
  * NTA Monitor Ltd.
  *
  * This program is free software; you can redistribute it and/or
@@ -16,7 +16,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *
- * $Id: arp-scan.c 10534 2007-04-10 10:17:13Z rsh $
+ * $Id: arp-scan.c 13669 2008-07-17 12:01:12Z rsh $
  *
  * arp-scan -- The ARP Scanner
  *
@@ -37,7 +37,7 @@
 
 #include "arp-scan.h"
 
-static char const rcsid[] = "$Id: arp-scan.c 10534 2007-04-10 10:17:13Z rsh $";   /* RCS ID for ident(1) */
+static char const rcsid[] = "$Id: arp-scan.c 13669 2008-07-17 12:01:12Z rsh $";   /* RCS ID for ident(1) */
 
 /* Global variables */
 static host_entry *helist = NULL;	/* Array of host entries */
@@ -49,6 +49,7 @@ static unsigned live_count;		/* Number of entries awaiting reply */
 static int verbose=0;			/* Verbose level */
 static int debug = 0;			/* Debug flag */
 static pcap_t *pcap_handle;		/* pcap handle */
+static pcap_dumper_t *pcap_dump_handle = NULL;	/* pcap savefile handle */
 static int pcap_fd;			/* Pcap file descriptor */
 static char filename[MAXLINE];		/* Target list file name */
 static int filename_flag=0;		/* Set if using target list file */
@@ -71,6 +72,7 @@ static int arp_sha_flag=0;		/* Source MAC address specified */
 static char ouifilename[MAXLINE];	/* OUI filename */
 static char iabfilename[MAXLINE];	/* IAB filename */
 static char macfilename[MAXLINE];	/* MAC filename */
+static char pcap_savefile[MAXLINE];	/* pcap savefile filename */
 static int arp_op=DEFAULT_ARP_OP;	/* ARP Operation code */
 static int arp_hrd=DEFAULT_ARP_HRD;	/* ARP hardware type */
 static int arp_pro=DEFAULT_ARP_PRO;	/* ARP protocol */
@@ -85,18 +87,18 @@ static unsigned char *padding=NULL;
 static size_t padding_len=0;
 static struct hash_control *hash_table;
 static int localnet_flag=0;		/* Scan local network */
+static int llc_flag=0;			/* Use 802.2 LLC with SNAP */
+static int ieee_8021q_vlan=-1;		/* Use 802.1Q VLAN tagging if >= 0 */
 
 int
 main(int argc, char *argv[]) {
-   char arg_str[MAXLINE];       /* Args as string for syslog */
    struct sockaddr_in sa_peer;
    struct timeval now;
-   unsigned char packet_in[MAXIP];      /* Received packet */
+   unsigned char packet_in[MAX_FRAME];      /* Received packet */
    struct timeval diff;         /* Difference between two timevals */
    int select_timeout;          /* Select timeout */
    ARP_UINT64 loop_timediff;    /* Time since last packet sent in us */
    ARP_UINT64 host_timediff; /* Time since last pkt sent to this host (us) */
-   int arg;
    struct timeval last_packet_time;     /* Time last packet was sent */
    int req_interval;            /* Requested per-packet interval */
    int cum_err=0;               /* Cumulative timing error */
@@ -121,15 +123,20 @@ main(int argc, char *argv[]) {
  *      Open syslog channel and log arguments if required.
  */
 #ifdef SYSLOG
-   openlog("arp-scan", LOG_PID, SYSLOG_FACILITY);
-   arg_str[0] = '\0';
-   for (arg=0; arg<argc; arg++) {
-      strlcat(arg_str, argv[arg], sizeof(arg_str));
-      if (arg < (argc-1)) {
-         strlcat(arg_str, " ", sizeof(arg_str));
+   {
+      char arg_str[MAXLINE];       /* Args as string for syslog */
+      int arg;
+
+      openlog("arp-scan", LOG_PID, SYSLOG_FACILITY);
+      arg_str[0] = '\0';
+      for (arg=0; arg<argc; arg++) {
+         strlcat(arg_str, argv[arg], sizeof(arg_str));
+         if (arg < (argc-1)) {
+            strlcat(arg_str, " ", sizeof(arg_str));
+         }
       }
+      info_syslog("Starting: %s", arg_str);
    }
-   info_syslog("Starting: %s", arg_str);
 #endif
 /*
  *      Initialise file names to the empty string.
@@ -137,6 +144,7 @@ main(int argc, char *argv[]) {
    ouifilename[0] = '\0';
    iabfilename[0] = '\0';
    macfilename[0] = '\0';
+   pcap_savefile[0] = '\0';
 /*
  *      Process options.
  */
@@ -149,14 +157,11 @@ main(int argc, char *argv[]) {
 /*
  *	Determine network interface to use.
  *	If the interface was specified with the --interface option then use
- *	that, otherwise if the environment variable "RMIF" exists then use
- *	that, failing that use pcap_lookupdev() to pick a suitable interface.
+ *	that, otherwise use pcap_lookupdev() to pick a suitable interface.
  */
    if (!if_name) { /* i/f not specified with --interface */
-      if (!(if_name=getenv("RMIF"))) {	/* No RMIF env var */
-         if (!(if_name=pcap_lookupdev(errbuf))) {
-            err_msg("pcap_lookupdev: %s", errbuf);
-         }
+      if (!(if_name=pcap_lookupdev(errbuf))) {
+         err_msg("pcap_lookupdev: %s", errbuf);
       }
    }
 /*
@@ -247,7 +252,15 @@ main(int argc, char *argv[]) {
          err_msg("ERROR: pcap_lookupnet: %s", errbuf);
       }
    }
-   filter_string=make_message("arp and ether dst %.2x:%.2x:%.2x:%.2x:%.2x:%.2x",
+/*
+ *	The filter string selects packets addresses to our interface address
+ *	that are either Ethernet-II ARP packets, 802.3 LLC/SNAP ARP packets
+ *	or 802.1Q tagged ARP packets.
+ */
+   filter_string=make_message("ether dst %.2x:%.2x:%.2x:%.2x:%.2x:%.2x and "
+                              "(arp or (ether[14:4]=0xaaaa0300 and "
+                              "ether[20:2]=0x0806) or (ether[12:2]=0x8100 and "
+                              "ether[16:2]=0x0806))",
                               interface_mac[0], interface_mac[1],
                               interface_mac[2], interface_mac[3],
                               interface_mac[4], interface_mac[5]);
@@ -265,6 +278,14 @@ main(int argc, char *argv[]) {
    if ((setuid(getuid())) < 0) {
       err_sys("setuid");
    }
+/*
+ *	Open pcap savefile is the --pcapsavefile (-W) option was specified
+ */
+   if (*pcap_savefile != '\0') {
+      if (!(pcap_dump_handle=pcap_dump_open(pcap_handle, pcap_savefile))) {
+         err_msg("pcap_dump_open: %s", pcap_geterr(pcap_handle));
+      }
+   }
 /*
  *      Check that the combination of specified options and arguments is
  *      valid.
@@ -549,7 +570,7 @@ main(int argc, char *argv[]) {
          if (debug) {print_times(); printf("main: Can't send packet yet.  loop_timediff=" ARP_UINT64_FORMAT "\n", loop_timediff);}
       } /* End If */
 
-      recvfrom_wto(pcap_fd, packet_in, MAXIP, (struct sockaddr *)&sa_peer,
+      recvfrom_wto(pcap_fd, packet_in, MAX_FRAME, (struct sockaddr *)&sa_peer,
                    select_timeout);
    } /* End While */
 
@@ -580,13 +601,13 @@ main(int argc, char *argv[]) {
  *
  *	Inputs:
  *
- *	n		The length of the received packet in bytes.
- *			Note that this can be more or less than the IP packet
- *			size because of minimum frame sizes or snaplength
- *			cutoff respectively.
- *	packet_in	The received packet
  *	he		The host entry corresponding to the received packet
- *	recv_addr	IP address that the packet was received from
+ *	arpei		ARP packet structure
+ *	extra_data	Extra data after ARP packet (padding)
+ *	extra_data_len	Length of extra data
+ *	framing		Framing type (e.g. Ethernet II, LLC)
+ *	vlan_id		802.1Q VLAN identifier, or -1 if not 802.1Q
+ *	frame_hdr	The Ethernet frame header
  *
  *      Returns:
  *
@@ -596,39 +617,39 @@ main(int argc, char *argv[]) {
  *      was received in the format: <IP-Address><TAB><Details>.
  */
 void
-display_packet(int n, const unsigned char *packet_in, host_entry *he,
-               struct in_addr *recv_addr) {
-   ether_hdr frame_hdr;
-   arp_ether_ipv4 arpei;
+display_packet(host_entry *he, arp_ether_ipv4 *arpei,
+               const unsigned char *extra_data, size_t extra_data_len,
+               int framing, int vlan_id, ether_hdr *frame_hdr) {
    char *msg;
    char *cp;
    char *cp2;
-   const unsigned char *ucp;
-   int extra_data;
    int nonzero=0;
 /*
- *	Set msg to the IP address of the host entry, plus the address of the
- *	responder if different, and a tab.
+ *	Set msg to the IP address of the host entry and a tab.
  */
    msg = make_message("%s\t", my_ntoa(he->addr));
-   if ((he->addr).s_addr != recv_addr->s_addr) {
-      cp = msg;
-      msg = make_message("%s(%s) ", cp, my_ntoa(*recv_addr));
-      free(cp);
-   }
-/*
- *      Unmarshal packet buffer into ARP structure
- */
-   unmarshal_arp_pkt(packet_in, &frame_hdr, &arpei);
 /*
  *	Decode ARP packet
  */
    cp = msg;
    msg = make_message("%s%.2x:%.2x:%.2x:%.2x:%.2x:%.2x", cp,
-                      arpei.ar_sha[0], arpei.ar_sha[1],
-                      arpei.ar_sha[2], arpei.ar_sha[3],
-                      arpei.ar_sha[4], arpei.ar_sha[5]);
+                      arpei->ar_sha[0], arpei->ar_sha[1],
+                      arpei->ar_sha[2], arpei->ar_sha[3],
+                      arpei->ar_sha[4], arpei->ar_sha[5]);
    free(cp);
+/*
+ *	Check that the source address in the Ethernet frame header is the same
+ *	as ar$sha in the ARP packet, and display the Ethernet source address
+ *	if it is different.
+ */
+   if ((memcmp(arpei->ar_sha, frame_hdr->src_addr, ETH_ALEN)) != 0) {
+      cp = msg;
+      msg = make_message("%s (%.2x:%.2x:%.2x:%.2x:%.2x:%.2x)", cp,
+                         frame_hdr->src_addr[0], frame_hdr->src_addr[1],
+                         frame_hdr->src_addr[2], frame_hdr->src_addr[3],
+                         frame_hdr->src_addr[4], frame_hdr->src_addr[5]);
+      free(cp);
+   }
 /*
  *	Find vendor from hash table and add to message if quiet if not in
  *	effect.
@@ -643,8 +664,8 @@ display_packet(int n, const unsigned char *packet_in, host_entry *he,
       int oui_end=12;
 
       snprintf(oui_string, 13, "%.2X%.2X%.2X%.2X%.2X%.2X",
-               arpei.ar_sha[0], arpei.ar_sha[1], arpei.ar_sha[2],
-               arpei.ar_sha[3], arpei.ar_sha[4], arpei.ar_sha[5]);
+               arpei->ar_sha[0], arpei->ar_sha[1], arpei->ar_sha[2],
+               arpei->ar_sha[3], arpei->ar_sha[4], arpei->ar_sha[5]);
       while (vendor == NULL && oui_end > 1) {
          oui_string[oui_end] = '\0';	/* Truncate oui string */
          vendor = hash_find(hash_table, oui_string);
@@ -660,24 +681,42 @@ display_packet(int n, const unsigned char *packet_in, host_entry *he,
  *	Check that any data after the ARP packet is zero.
  *	If it is non-zero, and verbose is selected, then print the padding.
  */
-   ucp = packet_in+ETHER_HDR_SIZE+ARP_PKT_SIZE;
-   extra_data = n-ETHER_HDR_SIZE-ARP_PKT_SIZE;
-   if (extra_data > 0) {
-      int i;
-      for (i=0; i<extra_data; i++) {
-         if (ucp[i] != '\0') {
-            nonzero=1;
-            break;
+      if (extra_data_len > 0) {
+         int i;
+         const unsigned char *ucp = extra_data;
+
+         for (i=0; i<extra_data_len; i++) {
+            if (ucp[i] != '\0') {
+               nonzero=1;
+               break;
+            }
          }
       }
-   }
-   if (nonzero && verbose) {
-      cp = msg;
-      cp2 = hexstring(ucp, (unsigned)extra_data);
-      msg = make_message("%s\tPadding=%s", cp, cp2);
-      free(cp2);
-      free(cp);
-   }
+      if (nonzero && verbose) {
+         cp = msg;
+         cp2 = hexstring(extra_data, extra_data_len);
+         msg = make_message("%s\tPadding=%s", cp, cp2);
+         free(cp2);
+         free(cp);
+      }
+/*
+ *	If the framing type is not Ethernet II, then report the framing type.
+ */
+      if (framing != FRAMING_ETHERNET_II) {
+         cp = msg;
+         if (framing == FRAMING_LLC_SNAP) {
+            msg = make_message("%s (802.2 LLC/SNAP)", cp);
+         }
+         free(cp);
+      }
+/*
+ *	If the packet uses 802.1Q VLAN tagging, report the VLAN ID.
+ */
+      if (vlan_id != -1) {
+         cp = msg;
+         msg = make_message("%s (802.1Q VLAN=%d)", cp, vlan_id);
+         free(cp);
+      }
 /*
  *      If the host entry is not live, then flag this as a duplicate.
  */
@@ -714,7 +753,7 @@ display_packet(int n, const unsigned char *packet_in, host_entry *he,
 int
 send_packet(link_t *link_handle, host_entry *he,
             struct timeval *last_packet_time) {
-   unsigned char buf[MAXIP];
+   unsigned char buf[MAX_FRAME];
    size_t buflen;
    ether_hdr frame_hdr;
    arp_ether_ipv4 arpei;
@@ -748,14 +787,7 @@ send_packet(link_t *link_handle, host_entry *he,
  *	Copy the required data into the output buffer "buf" and set "buflen"
  *	to the number of bytes in this buffer.
  */
-   marshal_arp_pkt(buf, &frame_hdr, &arpei, &buflen);
-/*
- *	Add padding if specified
- */
-   if (padding != NULL) {
-      memcpy(buf+buflen, padding, padding_len);
-      buflen += padding_len;
-   }
+   marshal_arp_pkt(buf, &frame_hdr, &arpei, &buflen, padding, padding_len);
 /*
  *	If he is NULL, just return with the packet length.
  */
@@ -812,6 +844,9 @@ clean_up(void) {
 
    printf("%u packets received by filter, %u packets dropped by kernel\n",
           stats.ps_recv, stats.ps_drop);
+   if (pcap_dump_handle) {
+      pcap_dump_close(pcap_dump_handle);
+   }
    pcap_close(pcap_handle);
 }
 
@@ -906,11 +941,9 @@ usage(int status) {
    fprintf(stderr, "\t\t\tlength includes the data-link header.\n");
    fprintf(stderr, "\t\t\tThe default is normally sufficient.\n");
    fprintf(stderr, "\n--interface=<i> or -I <i> Use network interface <i>.\n");
-   fprintf(stderr, "\t\t\tIf this option is not specified, the default is the\n");
-   fprintf(stderr, "\t\t\tvalue of the RMIF environment variable.  If RMIF is\n");
-   fprintf(stderr, "\t\t\tnot defined, then arp-scan will search the system\n");
-   fprintf(stderr, "\t\t\tinterface list for the lowest numbered, configured\n");
-   fprintf(stderr, "\t\t\tup interface (excluding loopback).\n");
+   fprintf(stderr, "\t\t\tIf this option is not specified, arp-scan will search\n");
+   fprintf(stderr, "\t\t\tthe system interface list for the lowest numbered,\n");
+   fprintf(stderr, "\t\t\tconfigured up interface (excluding loopback).\n");
    fprintf(stderr, "\t\t\tThe interface specified must support ARP.\n");
    fprintf(stderr, "\n--quiet or -q\t\tOnly display minimal output.\n");
    fprintf(stderr, "\t\t\tIf this option is specified, then only the minimum\n");
@@ -1013,6 +1046,25 @@ usage(int status) {
    fprintf(stderr, "\t\t\tPadding.  The default is no padding, although the\n");
    fprintf(stderr, "\t\t\tEthernet driver on the sending system may pad the\n");
    fprintf(stderr, "\t\t\tpacket to the minimum Ethernet frame length.\n");
+   fprintf(stderr, "\n--llc or -L\t\tUse RFC 1042 LLC framing with SNAP.\n");
+   fprintf(stderr, "\t\t\tThis option causes the outgoing ARP packets to use\n");
+   fprintf(stderr, "\t\t\tIEEE 802.2 framing with a SNAP header as described\n");
+   fprintf(stderr, "\t\t\tin RFC 1042.  The default is to use Ethernet-II\n");
+   fprintf(stderr, "\t\t\tframing.\n");
+   fprintf(stderr, "\t\t\tarp-scan will decode and display received ARP packets\n");
+   fprintf(stderr, "\t\t\tin either Ethernet-II or IEEE 802.2 formats\n");
+   fprintf(stderr, "\t\t\tirrespective of this option.\n");
+   fprintf(stderr, "\n--vlan=<v> or -Q <v>\tUse 802.1Q tagging with VLAN id <v>.\n");
+   fprintf(stderr, "\t\t\tThis option causes the outgoing ARP packets to use\n");
+   fprintf(stderr, "\t\t\t802.1Q VLAN tagging with a VLAN ID of <v>, which should\n");
+   fprintf(stderr, "\t\t\tbe in the range 0 to 4095 inclusive.\n");
+   fprintf(stderr, "\t\t\tarp-scan will always decode and display received ARP\n");
+   fprintf(stderr, "\t\t\tpackets in 802.1Q format irrespective of this option.\n");
+   fprintf(stderr, "\n--pcapsavefile=<p> or -W <p>\tWrite received packets to pcap savefile <p>.\n");
+   fprintf(stderr, "\t\t\tThis option causes received ARP responses to be written\n");
+   fprintf(stderr, "\t\t\tto a pcap savefile with the specified name.  This\n");
+   fprintf(stderr, "\t\t\tsavefile can be analyzed with programs that understand\n");
+   fprintf(stderr, "\t\t\tthe pcap file format, such as \"tcpdump\" and \"wireshark\".\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Report bugs or send suggestions to %s\n", PACKAGE_BUGREPORT);
    fprintf(stderr, "See the arp-scan homepage at http://www.nta-monitor.com/tools/arp-scan/\n");
@@ -1350,19 +1402,15 @@ advance_cursor(void) {
  *	he 	Pointer to the current position in the list.  Search runs
  *		backwards starting from this point.
  *	addr 	The source IP address that the packet came from.
- *	packet_in The received packet data.
- *	n	The length of the received packet.
  *
  *	Returns a pointer to the host entry associated with the specified IP
  *	or NULL if no match found.
  *
- *	This routine will normally find the host by IP address by comparing
- *	"addr" against "he->addr" for each entry in the list.  In this case,
- *	"packet_in" and "n" are not used.
+ *	This routine finds the host by IP address by comparing "addr" against
+ *	"he->addr" for each entry in the list.
  */
 host_entry *
-find_host(host_entry **he, struct in_addr *addr,
-          const unsigned char *packet_in, int n) {
+find_host(host_entry **he, struct in_addr *addr) {
    host_entry **p;
    int found = 0;
    unsigned iterations = 0;	/* Used for debugging */
@@ -1484,6 +1532,10 @@ callback(u_char *args, const struct pcap_pkthdr *header,
    int n = header->caplen;
    struct in_addr source_ip;
    host_entry *temp_cursor;
+   unsigned char extra_data[MAX_FRAME];
+   size_t extra_data_len;
+   int vlan_id;
+   int framing;
 /*
  *      Check that the packet is large enough to decode.
  */
@@ -1492,9 +1544,10 @@ callback(u_char *args, const struct pcap_pkthdr *header,
       return;
    }
 /*
- *	Unmarshal packet buffer into ARP structure
+ *	Unmarshal packet buffer into structures and determine framing type
  */
-   unmarshal_arp_pkt(packet_in, &frame_hdr, &arpei);
+   framing = unmarshal_arp_pkt(packet_in, n, &frame_hdr, &arpei, extra_data,
+                               &extra_data_len, &vlan_id);
 /*
  *	Determine source IP address.
  */
@@ -1507,7 +1560,7 @@ callback(u_char *args, const struct pcap_pkthdr *header,
  *	because we call advance_cursor() after sending each packet.  However,
  *	the time saved is minimal, and it's not worth the extra complexity.
  */
-   temp_cursor=find_host(cursor, &source_ip, packet_in, n);
+   temp_cursor=find_host(cursor, &source_ip);
    if (temp_cursor) {
 /*
  *	We found an IP match for the packet. 
@@ -1521,7 +1574,11 @@ callback(u_char *args, const struct pcap_pkthdr *header,
          warn_msg("---\tReceived packet #%u from %s",
                   temp_cursor->num_recv ,my_ntoa(source_ip));
       if ((temp_cursor->live || !ignore_dups)) {
-         display_packet(n, packet_in, temp_cursor, &source_ip);
+         if (pcap_dump_handle) {
+            pcap_dump((unsigned char *)pcap_dump_handle, header, packet_in);
+         }
+         display_packet(temp_cursor, &arpei, extra_data, extra_data_len,
+                        framing, vlan_id, &frame_hdr);
          responders++;
       }
       if (verbose > 1)
@@ -1585,10 +1642,13 @@ process_options(int argc, char *argv[]) {
       {"arptha", required_argument, 0, 'w'},
       {"srcaddr", required_argument, 0, 'S'},
       {"localnet", no_argument, 0, 'l'},
+      {"llc", no_argument, 0, 'L'},
+      {"vlan", required_argument, 0, 'Q'},
+      {"pcapsavefile", required_argument, 0, 'W'},
       {0, 0, 0, 0}
    };
    const char *short_options =
-      "f:hr:t:i:b:vVdn:I:qgRNB:O:s:o:H:p:T:P:a:A:y:u:w:S:F:m:l";
+      "f:hr:t:i:b:vVdn:I:qgRNB:O:s:o:H:p:T:P:a:A:y:u:w:S:F:m:lLQ:W:";
    int arg;
    int options_index=0;
 
@@ -1637,7 +1697,7 @@ process_options(int argc, char *argv[]) {
             debug++;
             break;
          case 'n':	/* --snap */
-            snaplen=strtol(optarg, (char **)NULL, 0);
+            snaplen=Strtol(optarg, 0);
             break;
          case 'I':	/* --interface */
             if_name = make_message("%s", optarg);
@@ -1685,13 +1745,13 @@ process_options(int argc, char *argv[]) {
             }
             break;
          case 'o':	/* --arpop */
-            arp_op=strtol(optarg, (char **)NULL, 0);
+            arp_op=Strtol(optarg, 0);
             break;
          case 'H':	/* --arphrd */
-            arp_hrd=strtol(optarg, (char **)NULL, 0);
+            arp_hrd=Strtol(optarg, 0);
             break;
          case 'p':	/* --arppro */
-            arp_pro=strtol(optarg, (char **)NULL, 0);
+            arp_pro=Strtol(optarg, 0);
             break;
          case 'T':	/* --destaddr */
             result = get_ether_addr(optarg, target_mac);
@@ -1699,10 +1759,10 @@ process_options(int argc, char *argv[]) {
                err_msg("Invalid target MAC address: %s", optarg);
             break;
          case 'P':	/* --arppln */
-            arp_pln=strtol(optarg, (char **)NULL, 0);
+            arp_pln=Strtol(optarg, 0);
             break;
          case 'a':	/* --arphln */
-            arp_hln=strtol(optarg, (char **)NULL, 0);
+            arp_hln=Strtol(optarg, 0);
             break;
          case 'A':	/* --padding */
             if (strlen(optarg) % 2)     /* Length is odd */
@@ -1710,7 +1770,7 @@ process_options(int argc, char *argv[]) {
             padding=hex2data(optarg, &padding_len);
             break;
          case 'y':	/* --prototype */
-            eth_pro=strtol(optarg, (char **)NULL, 0);
+            eth_pro=Strtol(optarg, 0);
             break;
          case 'u':	/* --arpsha */
             result = get_ether_addr(optarg, arp_sha);
@@ -1733,6 +1793,15 @@ process_options(int argc, char *argv[]) {
             localnet_flag = 1;
             numeric_flag=1;
             break;
+         case 'L':	/* --llc */
+            llc_flag = 1;
+            break;
+         case 'Q':	/* --vlan */
+            ieee_8021q_vlan = Strtol(optarg, 0);
+            break;
+         case 'W':	/* --pcapsavefile */
+            strlcpy(pcap_savefile, optarg, sizeof(pcap_savefile));
+            break;
          default:	/* Unknown option */
             usage(EXIT_FAILURE);
             break;	/* NOTREACHED */
@@ -1752,7 +1821,7 @@ process_options(int argc, char *argv[]) {
 void
 arp_scan_version (void) {
    fprintf(stderr, "%s\n\n", PACKAGE_STRING);
-   fprintf(stderr, "Copyright (C) 2005-2007 Roy Hills, NTA Monitor Ltd.\n");
+   fprintf(stderr, "Copyright (C) 2005-2008 Roy Hills, NTA Monitor Ltd.\n");
    fprintf(stderr, "arp-scan comes with NO WARRANTY to the extent permitted by law.\n");
    fprintf(stderr, "You may redistribute copies of arp-scan under the terms of the GNU\n");
    fprintf(stderr, "General Public License.\n");
@@ -1859,30 +1928,66 @@ my_ntoa(struct in_addr addr) {
  */
 void
 marshal_arp_pkt(unsigned char *buffer, ether_hdr *frame_hdr,
-                arp_ether_ipv4 *arp_pkt, size_t *buf_siz) {
+                arp_ether_ipv4 *arp_pkt, size_t *buf_siz,
+                const unsigned char *padding, size_t padding_len) {
+   unsigned char llc_snap[] = {0xAA, 0xAA, 0x03, 0x00, 0x00, 0x00, 0x08, 0x06};
+   unsigned char vlan_tag[] = {0x81, 0x00, 0x00, 0x00};
    unsigned char *cp;
+   size_t packet_size;
 
    cp = buffer;
-   *buf_siz = 0;
+/*
+ *	Set initial packet length to the size of an Ethernet frame using
+ *	Ethernet-II format plus the size of the ARP data.  This may be
+ *	increased later by LLC/SNAP frame format or padding after the
+ *	ARP data.
+ */
+   packet_size = sizeof(frame_hdr->dest_addr) + sizeof(frame_hdr->src_addr) +
+                 sizeof(frame_hdr->frame_type) +
+                 sizeof(arp_pkt->ar_hrd) + sizeof(arp_pkt->ar_pro) +
+                 sizeof(arp_pkt->ar_hln) + sizeof(arp_pkt->ar_pln) +
+                 sizeof(arp_pkt->ar_op)  + sizeof(arp_pkt->ar_sha) +
+                 sizeof(arp_pkt->ar_sip) + sizeof(arp_pkt->ar_tha) +
+                 sizeof(arp_pkt->ar_tip);
+/*
+ *	Copy the Ethernet frame header to the buffer.
+ */
+   memcpy(cp, &(frame_hdr->dest_addr), sizeof(frame_hdr->dest_addr));
+   cp += sizeof(frame_hdr->dest_addr);
+   memcpy(cp, &(frame_hdr->src_addr), sizeof(frame_hdr->src_addr));
+   cp += sizeof(frame_hdr->src_addr);
+/*
+ *	Add 802.1Q tag if we are using VLAN tagging
+ */
+   if (ieee_8021q_vlan != -1) {
+      uint16_t tci;
 
-   if (frame_hdr != NULL) {
-      *buf_siz += sizeof(frame_hdr->dest_addr) + sizeof(frame_hdr->src_addr) +
-                  sizeof(frame_hdr->frame_type);
-
-      memcpy(cp, &(frame_hdr->dest_addr), sizeof(frame_hdr->dest_addr));
-      cp += sizeof(frame_hdr->dest_addr);
-      memcpy(cp, &(frame_hdr->src_addr), sizeof(frame_hdr->src_addr));
-      cp += sizeof(frame_hdr->src_addr);
-      memcpy(cp, &(frame_hdr->frame_type), sizeof(frame_hdr->frame_type));
-      cp += sizeof(frame_hdr->frame_type);
+      tci = htons(ieee_8021q_vlan);
+      memcpy(cp, vlan_tag, sizeof(vlan_tag));
+      memcpy(cp+2, &tci, sizeof(tci));
+      cp += sizeof(vlan_tag);
+      packet_size += sizeof(vlan_tag);
    }
+   if (llc_flag) {	/* With 802.2 LLC framing, type field is frame size */
+      uint16_t frame_size;
 
-   *buf_siz += sizeof(arp_pkt->ar_hrd) + sizeof(arp_pkt->ar_pro) +
-               sizeof(arp_pkt->ar_hln) + sizeof(arp_pkt->ar_pln) +
-               sizeof(arp_pkt->ar_op)  + sizeof(arp_pkt->ar_sha) +
-               sizeof(arp_pkt->ar_sip) + sizeof(arp_pkt->ar_tha) +
-               sizeof(arp_pkt->ar_tip);
-
+      frame_size=htons(packet_size + sizeof(llc_snap));
+      memcpy(cp, &(frame_size), sizeof(frame_size));
+   } else {		/* Normal Ethernet-II framing */
+      memcpy(cp, &(frame_hdr->frame_type), sizeof(frame_hdr->frame_type));
+   }
+   cp += sizeof(frame_hdr->frame_type);
+/*
+ *	Add IEEE 802.2 LLC and SNAP fields if we are using LLC frame format.
+ */
+   if (llc_flag) {
+      memcpy(cp, llc_snap, sizeof(llc_snap));
+      cp += sizeof(llc_snap);
+      packet_size += sizeof(llc_snap);
+   }
+/*
+ *	Add the ARP data.
+ */
    memcpy(cp, &(arp_pkt->ar_hrd), sizeof(arp_pkt->ar_hrd));
    cp += sizeof(arp_pkt->ar_hrd);
    memcpy(cp, &(arp_pkt->ar_pro), sizeof(arp_pkt->ar_pro));
@@ -1900,6 +2005,22 @@ marshal_arp_pkt(unsigned char *buffer, ether_hdr *frame_hdr,
    memcpy(cp, &(arp_pkt->ar_tha), sizeof(arp_pkt->ar_tha));
    cp += sizeof(arp_pkt->ar_tha);
    memcpy(cp, &(arp_pkt->ar_tip), sizeof(arp_pkt->ar_tip));
+   cp += sizeof(arp_pkt->ar_tip);
+/*
+ *	Add padding if specified
+ */
+   if (padding != NULL) {
+      size_t safe_padding_len;
+
+      safe_padding_len = padding_len;
+      if (packet_size + padding_len > MAX_FRAME) {
+         safe_padding_len = MAX_FRAME - packet_size;
+      }
+      memcpy(cp, padding, safe_padding_len);
+      cp += safe_padding_len;
+      packet_size += safe_padding_len;
+   }
+   *buf_siz = packet_size;
 }
 
 /*
@@ -1908,17 +2029,31 @@ marshal_arp_pkt(unsigned char *buffer, ether_hdr *frame_hdr,
  *	Inputs:
  *
  *	buffer		Pointer to the input buffer
+ *	buf_len		Length of input buffer
  *	frame_hdr	The ethernet frame header
  *	arp_pkt		The arp packet data
+ *	extra_data	Any extra data after the ARP data (typically padding)
+ *	extra_data_len	Length of extra data
+ *	vlan_id		802.1Q VLAN identifier
  *
  *	Returns:
  *
- *	None
+ *	An integer representing the data link framing:
+ *	0 = Ethernet-II
+ *	1 = 802.3 with LLC/SNAP
+ *
+ *	extra_data and extra_data_len are only calculated and returned if
+ *	extra_data is not NULL.
+ *
+ *	vlan_id is set to -1 if the packet does not use 802.1Q tagging.
  */
-void
-unmarshal_arp_pkt(const unsigned char *buffer, ether_hdr *frame_hdr,
-                  arp_ether_ipv4 *arp_pkt) {
+int
+unmarshal_arp_pkt(const unsigned char *buffer, size_t buf_len,
+                  ether_hdr *frame_hdr, arp_ether_ipv4 *arp_pkt,
+                  unsigned char *extra_data, size_t *extra_data_len,
+                  int *vlan_id) {
    const unsigned char *cp;
+   int framing=FRAMING_ETHERNET_II;
 
    cp = buffer;
 /*
@@ -1928,8 +2063,31 @@ unmarshal_arp_pkt(const unsigned char *buffer, ether_hdr *frame_hdr,
    cp += sizeof(frame_hdr->dest_addr);
    memcpy(&(frame_hdr->src_addr), cp, sizeof(frame_hdr->src_addr));
    cp += sizeof(frame_hdr->src_addr);
+/*
+ *	Check for 802.1Q VLAN tagging, indicated by a type code of
+ *	0x8100 (TPID).
+ */
+   if (*cp == 0x81 && *(cp+1) == 0x00) {
+      uint16_t tci;
+      cp += 2;	/* Skip TPID */
+      memcpy(&tci, cp, sizeof(tci));
+      cp += 2;	/* Skip TCI */
+      *vlan_id = ntohs(tci);
+      *vlan_id &= 0x0fff;	/* Mask off PRI and CFI */
+   } else {
+      *vlan_id = -1;
+   }
    memcpy(&(frame_hdr->frame_type), cp, sizeof(frame_hdr->frame_type));
    cp += sizeof(frame_hdr->frame_type);
+/*
+ *	Check for an LLC header with SNAP.  If this is present, the 802.2 LLC
+ *	header will contain DSAP=0xAA, SSAP=0xAA, Control=0x03.
+ *	If this 802.2 LLC header is present, skip it and the SNAP header
+ */
+   if (*cp == 0xAA && *(cp+1) == 0xAA && *(cp+2) == 0x03) {
+      cp += 8;	/* Skip eight bytes */
+      framing = FRAMING_LLC_SNAP;
+   }
 /*
  *	Extract the ARP packet data
  */
@@ -1950,6 +2108,19 @@ unmarshal_arp_pkt(const unsigned char *buffer, ether_hdr *frame_hdr,
    memcpy(&(arp_pkt->ar_tha), cp, sizeof(arp_pkt->ar_tha));
    cp += sizeof(arp_pkt->ar_tha);
    memcpy(&(arp_pkt->ar_tip), cp, sizeof(arp_pkt->ar_tip));
+   cp += sizeof(arp_pkt->ar_tip);
+
+   if (extra_data != NULL) {
+      int length;
+
+      length = buf_len - (cp - buffer);
+      if (length > 0) {		/* Extra data after ARP packet */
+         memcpy(extra_data, cp, length);
+      }
+      *extra_data_len = length;
+   }
+
+   return framing;
 }
 
 /*
